@@ -31,7 +31,7 @@ class ConfigError(DvcException):
     """DVC config exception."""
 
     def __init__(self, msg):
-        super().__init__("config file error: {}".format(msg))
+        super().__init__(f"config file error: {msg}")
 
 
 class NoRemoteError(ConfigError):
@@ -89,7 +89,7 @@ def ByUrl(mapping):
         if os.name == "nt" and len(parsed.scheme) == 1 and parsed.netloc == "":
             return schemas[""](data)
         if parsed.scheme not in schemas:
-            raise Invalid("Unsupported URL type {}://".format(parsed.scheme))
+            raise Invalid(f"Unsupported URL type {parsed.scheme}://")
 
         return schemas[parsed.scheme](data)
 
@@ -151,6 +151,7 @@ SCHEMA = {
                     Optional("listobjects", default=False): Bool,
                     Optional("use_ssl", default=True): Bool,
                     "sse": str,
+                    "sse_kms_key_id": str,
                     "acl": str,
                     "grant_read": str,
                     "grant_read_acp": str,
@@ -289,7 +290,7 @@ class Config(dict):
         Raises:
             ConfigError: thrown if config has an invalid format.
         """
-        conf = self._load_config_to_level()
+        conf = self.load_config_to_level()
 
         if validate:
             conf = self.validate(conf)
@@ -349,7 +350,7 @@ class Config(dict):
         }
         return Schema(dirs_schema, extra=ALLOW_EXTRA)(conf)
 
-    def _load_config_to_level(self, level=None):
+    def load_config_to_level(self, level=None):
         merged_conf = {}
         for merge_level in self.LEVELS:
             if merge_level == level:
@@ -368,7 +369,7 @@ class Config(dict):
 
         conf = self._save_paths(conf, self.files[level])
 
-        merged_conf = self._load_config_to_level(level)
+        merged_conf = self.load_config_to_level(level)
         _merge(merged_conf, conf)
         self.validate(merged_conf)
 
@@ -389,7 +390,7 @@ def _load_config(filename):
 
 
 def _save_config(filename, conf_dict):
-    logger.debug("Writing '{}'.".format(filename))
+    logger.debug(f"Writing '{filename}'.")
     os.makedirs(os.path.dirname(filename), exist_ok=True)
 
     config = configobj.ConfigObj(_pack_remotes(conf_dict))
@@ -416,7 +417,7 @@ def _pack_remotes(conf):
 
     # Transform remote.name -> 'remote "name"'
     for name, val in conf["remote"].items():
-        result['remote "{}"'.format(name)] = val
+        result[f'remote "{name}"'] = val
     result.pop("remote", None)
 
     return result
