@@ -2,120 +2,101 @@ import errno
 import logging
 import os
 import stat
-from concurrent.futures import as_completed, ThreadPoolExecutor
-from functools import partial
+from concurrent.futures import ThreadPoolExecutor, as_completed
+from functools import partial, wraps
 
 from funcy import cached_property, concat
-
 from shortuuid import uuid
 
-from dvc.compat import fspath_py35
-from dvc.exceptions import DvcException, DownloadError, UploadError
+from dvc.exceptions import DownloadError, DvcException, UploadError
 from dvc.path_info import PathInfo
 from dvc.progress import Tqdm
 from dvc.remote.base import (
-    index_locked,
-    BaseRemote,
-    STATUS_MAP,
     STATUS_DELETED,
+    STATUS_MAP,
     STATUS_MISSING,
     STATUS_NEW,
+    BaseRemoteTree,
+    CloudCache,
+    Remote,
+    index_locked,
 )
 from dvc.remote.index import RemoteIndexNoop
 from dvc.scheme import Schemes
-from dvc.scm.tree import is_working_tree
+from dvc.scm.tree import WorkingTree, is_working_tree
 from dvc.system import System
 from dvc.utils import file_md5, relpath, tmp_fname
-from dvc.utils.fs import copyfile, move, makedirs, remove, walk_files
+from dvc.utils.fs import (
+    copy_fobj_to_file,
+    copyfile,
+    makedirs,
+    move,
+    remove,
+    walk_files,
+)
 
 logger = logging.getLogger(__name__)
 
 
-class LocalRemote(BaseRemote):
+class LocalRemoteTree(BaseRemoteTree):
     scheme = Schemes.LOCAL
-    path_cls = PathInfo
+    PATH_CLS = PathInfo
     PARAM_CHECKSUM = "md5"
     PARAM_PATH = "path"
     TRAVERSE_PREFIX_LEN = 2
-    INDEX_CLS = RemoteIndexNoop
-
     UNPACKED_DIR_SUFFIX = ".unpacked"
 
-    DEFAULT_CACHE_TYPES = ["reflink", "copy"]
-
     CACHE_MODE = 0o444
     SHARED_MODE_MAP = {None: (0o644, 0o755), "group": (0o664, 0o775)}
 
     def __init__(self, repo, config):
         super().__init__(repo, config)
-        self.cache_dir = config.get("url")
-        self._dir_info = {}
+        url = config.get("url")
+        self.path_info = self.PATH_CLS(url) if url else None
 
     @property
     def state(self):
-        return self.repo.state
+        from dvc.state import StateNoop
 
-    @property
-    def cache_dir(self):
-        return self.path_info.fspath if self.path_info else None
-
-    @cache_dir.setter
-    def cache_dir(self, value):
-        self.path_info = PathInfo(value) if value else None
-
-    @classmethod
-    def supported(cls, config):
-        return True
+        return self.repo.state if self.repo else StateNoop()
 
     @cached_property
-    def cache_path(self):
-        return os.path.abspath(self.cache_dir)
+    def work_tree(self):
+        # When using repo.brancher, repo.tree may change to/from WorkingTree to
+        # GitTree arbitarily. When repo.tree is GitTree, local cache needs to
+        # use its own WorkingTree instance.
+        if self.repo:
+            return WorkingTree(self.repo.root_dir)
+        return None
 
-    def checksum_to_path(self, checksum):
-        return os.path.join(self.cache_path, checksum[0:2], checksum[2:])
-
-    def list_cache_paths(self, prefix=None, progress_callback=None):
-        assert self.path_info is not None
-        if prefix:
-            path_info = self.path_info / prefix[:2]
-        else:
-            path_info = self.path_info
-        if progress_callback:
-            for path in walk_files(path_info):
-                progress_callback()
-                yield path
-        else:
-            yield from walk_files(path_info)
-
-    def get(self, md5):
-        if not md5:
-            return None
-
-        return self.checksum_to_path_info(md5).url
+    @staticmethod
+    def open(path_info, mode="r", encoding=None):
+        return open(path_info, mode=mode, encoding=encoding)
 
     def exists(self, path_info):
-        assert is_working_tree(self.repo.tree)
         assert isinstance(path_info, str) or path_info.scheme == "local"
-        return self.repo.tree.exists(fspath_py35(path_info))
+        if not self.repo:
+            return os.path.exists(path_info)
+        return self.work_tree.exists(path_info)
 
-    def makedirs(self, path_info):
-        makedirs(path_info, exist_ok=True, mode=self._dir_mode)
+    def isfile(self, path_info):
+        if not self.repo:
+            return os.path.isfile(path_info)
+        return self.work_tree.isfile(path_info)
 
-    def already_cached(self, path_info):
-        assert path_info.scheme in ["", "local"]
+    def isdir(self, path_info):
+        if not self.repo:
+            return os.path.isdir(path_info)
+        return self.work_tree.isdir(path_info)
 
-        current_md5 = self.get_checksum(path_info)
+    def iscopy(self, path_info):
+        return not (
+            System.is_symlink(path_info) or System.is_hardlink(path_info)
+        )
 
-        if not current_md5:
-            return False
-
-        return not self.changed_cache(current_md5)
-
-    def _verify_link(self, path_info, link_type):
-        if link_type == "hardlink" and self.getsize(path_info) == 0:
-            return
-
-        super()._verify_link(path_info, link_type)
+    def walk_files(self, path_info, **kwargs):
+        for fname in self.work_tree.walk_files(path_info):
+            yield PathInfo(fname)
 
     def is_empty(self, path_info):
         path = path_info.fspath
@@ -128,32 +109,6 @@ class LocalRemote(BaseRemote):
 
         return False
 
-    @staticmethod
-    def isfile(path_info):
-        return os.path.isfile(fspath_py35(path_info))
-
-    @staticmethod
-    def isdir(path_info):
-        return os.path.isdir(fspath_py35(path_info))
-
-    def iscopy(self, path_info):
-        return not (
-            System.is_symlink(path_info) or System.is_hardlink(path_info)
-        )
-
-    @staticmethod
-    def getsize(path_info):
-        return os.path.getsize(fspath_py35(path_info))
-
-    def walk_files(self, path_info):
-        assert is_working_tree(self.repo.tree)
-
-        for fname in self.repo.tree.walk_files(path_info):
-            yield PathInfo(fname)
-
-    def get_file_checksum(self, path_info):
-        return file_md5(path_info)[0]
-
     def remove(self, path_info):
         if isinstance(path_info, PathInfo):
             if path_info.scheme != "local":
@@ -165,6 +120,9 @@ class LocalRemote(BaseRemote):
         if self.exists(path):
             remove(path)
 
+    def makedirs(self, path_info):
+        makedirs(path_info, exist_ok=True, mode=self.dir_mode)
+
     def move(self, from_info, to_info, mode=None):
         if from_info.scheme != "local" or to_info.scheme != "local":
             raise NotImplementedError
@@ -173,9 +131,9 @@ class LocalRemote(BaseRemote):
 
         if mode is None:
             if self.isfile(from_info):
-                mode = self._file_mode
+                mode = self.file_mode
             else:
-                mode = self._dir_mode
+                mode = self.dir_mode
 
         move(from_info, to_info, mode=mode)
 
@@ -183,8 +141,19 @@ class LocalRemote(BaseRemote):
         tmp_info = to_info.parent / tmp_fname(to_info.name)
         try:
             System.copy(from_info, tmp_info)
-            os.chmod(fspath_py35(tmp_info), self._file_mode)
-            os.rename(fspath_py35(tmp_info), fspath_py35(to_info))
+            os.chmod(tmp_info, self.file_mode)
+            os.rename(tmp_info, to_info)
+        except Exception:
+            self.remove(tmp_info)
+            raise
+
+    def copy_fobj(self, fobj, to_info):
+        self.makedirs(to_info.parent)
+        tmp_info = to_info.parent / tmp_fname(to_info.name)
+        try:
+            copy_fobj_to_file(fobj, tmp_info)
+            os.chmod(tmp_info, self.file_mode)
+            os.rename(tmp_info, to_info)
         except Exception:
             self.remove(tmp_info)
             raise
@@ -202,7 +171,7 @@ class LocalRemote(BaseRemote):
         # and the cache type is `hardlink`, we might reach link limits and
         # will get something like: `too many links error`
         #
-        # This is because all those empty files will have the same checksum
+        # This is because all those empty files will have the same hash
         # (i.e. 68b329da9893e34099c7d8ad5cb9c940), therefore, they will be
         # linked to the same file in the cache.
         #
@@ -232,20 +201,81 @@ class LocalRemote(BaseRemote):
         System.reflink(from_info, tmp_info)
         # NOTE: reflink has its own separate inode, so you can set permissions
         # that are different from the source.
-        os.chmod(fspath_py35(tmp_info), self._file_mode)
-        os.rename(fspath_py35(tmp_info), fspath_py35(to_info))
+        os.chmod(tmp_info, self.file_mode)
+        os.rename(tmp_info, to_info)
 
-    def cache_exists(self, checksums, jobs=None, name=None):
-        return [
-            checksum
-            for checksum in Tqdm(
-                checksums,
-                unit="file",
-                desc="Querying "
-                + ("cache in " + name if name else "local cache"),
+    def _unprotect_file(self, path):
+        if System.is_symlink(path) or System.is_hardlink(path):
+            logger.debug(f"Unprotecting '{path}'")
+            tmp = os.path.join(os.path.dirname(path), "." + uuid())
+
+            # The operations order is important here - if some application
+            # would access the file during the process of copyfile then it
+            # would get only the part of file. So, at first, the file should be
+            # copied with the temporary name, and then original file should be
+            # replaced by new.
+            copyfile(path, tmp, name="Unprotecting '{}'".format(relpath(path)))
+            remove(path)
+            os.rename(tmp, path)
+
+        else:
+            logger.debug(
+                "Skipping copying for '{}', since it is not "
+                "a symlink or a hardlink.".format(path)
             )
-            if not self.changed_cache_file(checksum)
-        ]
+
+        os.chmod(path, self.file_mode)
+
+    def _unprotect_dir(self, path):
+        assert is_working_tree(self.repo.tree)
+
+        for fname in self.repo.tree.walk_files(path):
+            self._unprotect_file(fname)
+
+    def unprotect(self, path_info):
+        path = path_info.fspath
+        if not os.path.exists(path):
+            raise DvcException(f"can't unprotect non-existing data '{path}'")
+
+        if os.path.isdir(path):
+            self._unprotect_dir(path)
+        else:
+            self._unprotect_file(path)
+
+    def protect(self, path_info):
+        path = os.fspath(path_info)
+        mode = self.CACHE_MODE
+
+        try:
+            os.chmod(path, mode)
+        except OSError as exc:
+            # There is nothing we need to do in case of a read-only file system
+            if exc.errno == errno.EROFS:
+                return
+
+            # In shared cache scenario, we might not own the cache file, so we
+            # need to check if cache file is already protected.
+            if exc.errno not in [errno.EPERM, errno.EACCES]:
+                raise
+
+            actual = stat.S_IMODE(os.stat(path).st_mode)
+            if actual != mode:
+                raise
+
+    def is_protected(self, path_info):
+        try:
+            mode = os.stat(path_info).st_mode
+        except FileNotFoundError:
+            return False
+
+        return stat.S_IMODE(mode) == self.CACHE_MODE
+
+    def get_file_hash(self, path_info):
+        return file_md5(path_info)[0]
+
+    @staticmethod
+    def getsize(path_info):
+        return os.path.getsize(path_info)
 
     def _upload(
         self, from_file, to_info, name=None, no_progress_bar=False, **_kwargs
@@ -256,18 +286,122 @@ class LocalRemote(BaseRemote):
         copyfile(
             from_file, tmp_file, name=name, no_progress_bar=no_progress_bar
         )
-        os.rename(tmp_file, fspath_py35(to_info))
 
+        self.protect(tmp_file)
+        os.rename(tmp_file, to_info)
+
+    @staticmethod
     def _download(
-        self, from_info, to_file, name=None, no_progress_bar=False, **_kwargs
+        from_info, to_file, name=None, no_progress_bar=False, **_kwargs
     ):
         copyfile(
             from_info, to_file, no_progress_bar=no_progress_bar, name=name
         )
 
-    @staticmethod
-    def open(path_info, mode="r", encoding=None):
-        return open(fspath_py35(path_info), mode=mode, encoding=encoding)
+    def list_paths(self, prefix=None, progress_callback=None):
+        assert self.path_info is not None
+        if prefix:
+            path_info = self.path_info / prefix[:2]
+            if not self.exists(path_info):
+                return
+        else:
+            path_info = self.path_info
+        # NOTE: use utils.fs walk_files since tree.walk_files will not follow
+        # symlinks
+        if progress_callback:
+            for path in walk_files(path_info):
+                progress_callback()
+                yield path
+        else:
+            yield from walk_files(path_info)
+
+    def _remove_unpacked_dir(self, hash_):
+        info = self.hash_to_path_info(hash_)
+        path_info = info.with_name(info.name + self.UNPACKED_DIR_SUFFIX)
+        self.remove(path_info)
+
+
+def _log_exceptions(func, operation):
+    @wraps(func)
+    def wrapper(from_info, to_info, *args, **kwargs):
+        try:
+            func(from_info, to_info, *args, **kwargs)
+            return 0
+        except Exception as exc:
+            # NOTE: this means we ran out of file descriptors and there is no
+            # reason to try to proceed, as we will hit this error anyways.
+            if isinstance(exc, OSError) and exc.errno == errno.EMFILE:
+                raise
+
+            logger.exception(
+                "failed to %s '%s' to '%s'", operation, from_info, to_info
+            )
+            return 1
+
+    return wrapper
+
+
+class LocalRemote(Remote):
+    INDEX_CLS = RemoteIndexNoop
+
+
+class LocalCache(CloudCache):
+    DEFAULT_CACHE_TYPES = ["reflink", "copy"]
+    CACHE_MODE = LocalRemoteTree.CACHE_MODE
+
+    def __init__(self, tree):
+        super().__init__(tree)
+        self.cache_dir = tree.config.get("url")
+
+    @property
+    def cache_dir(self):
+        return self.tree.path_info.fspath if self.tree.path_info else None
+
+    @cache_dir.setter
+    def cache_dir(self, value):
+        self.tree.path_info = PathInfo(value) if value else None
+
+    @classmethod
+    def supported(cls, config):
+        return True
+
+    @cached_property
+    def cache_path(self):
+        return os.path.abspath(self.cache_dir)
+
+    def hash_to_path(self, hash_):
+        # NOTE: `self.cache_path` is already normalized so we can simply use
+        # `os.sep` instead of `os.path.join`. This results in this helper
+        # being ~5.5 times faster.
+        return f"{self.cache_path}{os.sep}{hash_[0:2]}{os.sep}{hash_[2:]}"
+
+    def hashes_exist(self, hashes, jobs=None, name=None):
+        return [
+            hash_
+            for hash_ in Tqdm(
+                hashes,
+                unit="file",
+                desc="Querying "
+                + ("cache in " + name if name else "local cache"),
+            )
+            if not self.changed_cache_file(hash_)
+        ]
+
+    def already_cached(self, path_info):
+        assert path_info.scheme in ["", "local"]
+
+        current_md5 = self.get_hash(path_info)
+
+        if not current_md5:
+            return False
+
+        return not self.changed_cache(current_md5)
+
+    def _verify_link(self, path_info, link_type):
+        if link_type == "hardlink" and self.tree.getsize(path_info) == 0:
+            return
+
+        super()._verify_link(path_info, link_type)
 
     @index_locked
     def status(
@@ -300,17 +434,15 @@ class LocalRemote(BaseRemote):
 
         dir_status_info contains status for .dir files, file_status_info
         contains status for all other files, and dir_contents is a dict of
-        {dir_checksum: set(file_checksum, ...)} which can be used to map
+        {dir_hash: set(file_hash, ...)} which can be used to map
         a .dir file to its file contents.
         """
-        logger.debug(
-            "Preparing to collect status from {}".format(remote.path_info)
-        )
+        logger.debug(f"Preparing to collect status from {remote.path_info}")
         md5s = set(named_cache.scheme_keys(self.scheme))
 
         logger.debug("Collecting information from local cache...")
         local_exists = frozenset(
-            self.cache_exists(md5s, jobs=jobs, name=self.cache_dir)
+            self.hashes_exist(md5s, jobs=jobs, name=self.cache_dir)
         )
 
         # This is a performance optimization. We can safely assume that,
@@ -325,12 +457,12 @@ class LocalRemote(BaseRemote):
             dir_md5s = set(named_cache.dir_keys(self.scheme))
             if dir_md5s:
                 remote_exists.update(
-                    self._indexed_dir_checksums(named_cache, remote, dir_md5s)
+                    self._indexed_dir_hashes(named_cache, remote, dir_md5s)
                 )
                 md5s.difference_update(remote_exists)
             if md5s:
                 remote_exists.update(
-                    remote.cache_exists(
+                    remote.hashes_exist(
                         md5s, jobs=jobs, name=str(remote.path_info)
                     )
                 )
@@ -341,23 +473,23 @@ class LocalRemote(BaseRemote):
     def _make_status(
         self, named_cache, show_checksums, local_exists, remote_exists
     ):
-        def make_names(checksum, names):
-            return {"name": checksum if show_checksums else " ".join(names)}
+        def make_names(hash_, names):
+            return {"name": hash_ if show_checksums else " ".join(names)}
 
         dir_status = {}
         file_status = {}
         dir_contents = {}
-        for checksum, item in named_cache[self.scheme].items():
+        for hash_, item in named_cache[self.scheme].items():
             if item.children:
-                dir_status[checksum] = make_names(checksum, item.names)
-                dir_contents[checksum] = set()
-                for child_checksum, child in item.children.items():
-                    file_status[child_checksum] = make_names(
-                        child_checksum, child.names
+                dir_status[hash_] = make_names(hash_, item.names)
+                dir_contents[hash_] = set()
+                for child_hash, child in item.children.items():
+                    file_status[child_hash] = make_names(
+                        child_hash, child.names
                     )
-                    dir_contents[checksum].add(child_checksum)
+                    dir_contents[hash_].add(child_hash)
             else:
-                file_status[checksum] = make_names(checksum, item.names)
+                file_status[hash_] = make_names(hash_, item.names)
 
         self._fill_statuses(dir_status, local_exists, remote_exists)
         self._fill_statuses(file_status, local_exists, remote_exists)
@@ -366,50 +498,50 @@ class LocalRemote(BaseRemote):
 
         return dir_status, file_status, dir_contents
 
-    def _indexed_dir_checksums(self, named_cache, remote, dir_md5s):
-        # Validate our index by verifying all indexed .dir checksums
+    def _indexed_dir_hashes(self, named_cache, remote, dir_md5s):
+        # Validate our index by verifying all indexed .dir hashes
         # still exist on the remote
-        indexed_dirs = set(remote.index.dir_checksums())
+        indexed_dirs = set(remote.index.dir_hashes())
         indexed_dir_exists = set()
         if indexed_dirs:
             indexed_dir_exists.update(
-                remote._cache_object_exists(indexed_dirs)
+                remote.tree.list_hashes_exists(indexed_dirs)
             )
             missing_dirs = indexed_dirs.difference(indexed_dir_exists)
             if missing_dirs:
                 logger.debug(
-                    "Remote cache missing indexed .dir checksums '{}', "
+                    "Remote cache missing indexed .dir hashes '{}', "
                     "clearing remote index".format(", ".join(missing_dirs))
                 )
                 remote.index.clear()
 
-        # Check if non-indexed (new) dir checksums exist on remote
+        # Check if non-indexed (new) dir hashes exist on remote
         dir_exists = dir_md5s.intersection(indexed_dir_exists)
-        dir_exists.update(remote._cache_object_exists(dir_md5s - dir_exists))
+        dir_exists.update(
+            remote.tree.list_hashes_exists(dir_md5s - dir_exists)
+        )
 
-        # If .dir checksum exists on the remote, assume directory contents
+        # If .dir hash exists on the remote, assume directory contents
         # still exists on the remote
-        for dir_checksum in dir_exists:
-            file_checksums = list(
-                named_cache.child_keys(self.scheme, dir_checksum)
-            )
-            if dir_checksum not in remote.index:
+        for dir_hash in dir_exists:
+            file_hashes = list(named_cache.child_keys(self.scheme, dir_hash))
+            if dir_hash not in remote.index:
                 logger.debug(
                     "Indexing new .dir '{}' with '{}' nested files".format(
-                        dir_checksum, len(file_checksums)
+                        dir_hash, len(file_hashes)
                     )
                 )
-                remote.index.update([dir_checksum], file_checksums)
-            yield dir_checksum
-            yield from file_checksums
+                remote.index.update([dir_hash], file_hashes)
+            yield dir_hash
+            yield from file_hashes
 
     @staticmethod
-    def _fill_statuses(checksum_info_dir, local_exists, remote_exists):
+    def _fill_statuses(hash_info_dir, local_exists, remote_exists):
         # Using sets because they are way faster for lookups
         local = set(local_exists)
         remote = set(remote_exists)
 
-        for md5, info in checksum_info_dir.items():
+        for md5, info in hash_info_dir.items():
             status = STATUS_MAP[(md5 in local, md5 in remote)]
             info["status"] = status
 
@@ -417,15 +549,15 @@ class LocalRemote(BaseRemote):
         cache = []
         path_infos = []
         names = []
-        checksums = []
+        hashes = []
         for md5, info in Tqdm(
             status_info.items(), desc="Analysing status", unit="file"
         ):
             if info["status"] == status:
-                cache.append(self.checksum_to_path_info(md5))
-                path_infos.append(remote.checksum_to_path_info(md5))
+                cache.append(self.hash_to_path_info(md5))
+                path_infos.append(remote.hash_to_path_info(md5))
                 names.append(info["name"])
-                checksums.append(md5)
+                hashes.append(md5)
 
         if download:
             to_infos = cache
@@ -434,7 +566,7 @@ class LocalRemote(BaseRemote):
             to_infos = path_infos
             from_infos = cache
 
-        return from_infos, to_infos, names, checksums
+        return from_infos, to_infos, names, hashes
 
     def _process(
         self,
@@ -453,19 +585,19 @@ class LocalRemote(BaseRemote):
 
         if download:
             func = partial(
-                remote.download,
-                dir_mode=self._dir_mode,
-                file_mode=self._file_mode,
+                _log_exceptions(remote.tree.download, "download"),
+                dir_mode=self.tree.dir_mode,
+                file_mode=self.tree.file_mode,
             )
             status = STATUS_DELETED
             desc = "Downloading"
         else:
-            func = remote.upload
+            func = _log_exceptions(remote.tree.upload, "upload")
             status = STATUS_NEW
             desc = "Uploading"
 
         if jobs is None:
-            jobs = remote.JOBS
+            jobs = remote.tree.JOBS
 
         dir_status, file_status, dir_contents = self._status(
             named_cache,
@@ -486,26 +618,28 @@ class LocalRemote(BaseRemote):
             func = pbar.wrap_fn(func)
             with ThreadPoolExecutor(max_workers=jobs) as executor:
                 if download:
-                    fails = sum(executor.map(func, *dir_plans))
-                    fails += sum(executor.map(func, *file_plans))
+                    from_infos, to_infos, names, _ = (
+                        d + f for d, f in zip(dir_plans, file_plans)
+                    )
+                    fails = sum(
+                        executor.map(func, from_infos, to_infos, names)
+                    )
                 else:
                     # for uploads, push files first, and any .dir files last
 
                     file_futures = {}
-                    for from_info, to_info, name, checksum in zip(*file_plans):
-                        file_futures[checksum] = executor.submit(
+                    for from_info, to_info, name, hash_ in zip(*file_plans):
+                        file_futures[hash_] = executor.submit(
                             func, from_info, to_info, name
                         )
                     dir_futures = {}
-                    for from_info, to_info, name, dir_checksum in zip(
-                        *dir_plans
-                    ):
+                    for from_info, to_info, name, dir_hash in zip(*dir_plans):
                         wait_futures = {
                             future
-                            for file_checksum, future in file_futures.items()
-                            if file_checksum in dir_contents[dir_checksum]
+                            for file_hash, future in file_futures.items()
+                            if file_hash in dir_contents[dir_hash]
                         }
-                        dir_futures[dir_checksum] = executor.submit(
+                        dir_futures[dir_hash] = executor.submit(
                             self._dir_upload,
                             func,
                             wait_futures,
@@ -528,16 +662,14 @@ class LocalRemote(BaseRemote):
 
         if not download:
             # index successfully pushed dirs
-            for dir_checksum, future in dir_futures.items():
+            for dir_hash, future in dir_futures.items():
                 if future.result() == 0:
-                    file_checksums = dir_contents[dir_checksum]
+                    file_hashes = dir_contents[dir_hash]
                     logger.debug(
                         "Indexing pushed dir '{}' with "
-                        "'{}' nested files".format(
-                            dir_checksum, len(file_checksums)
-                        )
+                        "'{}' nested files".format(dir_hash, len(file_hashes))
                     )
-                    remote.index.update([dir_checksum], file_checksums)
+                    remote.index.update([dir_hash], file_hashes)
 
         return len(dir_plans[0]) + len(file_plans[0])
 
@@ -551,9 +683,7 @@ class LocalRemote(BaseRemote):
                     "failed to upload full contents of '{}', "
                     "aborting .dir file upload".format(name)
                 )
-                logger.error(
-                    "failed to upload '{}' to '{}'".format(from_info, to_info)
-                )
+                logger.error(f"failed to upload '{from_info}' to '{to_info}'")
                 return 1
         return func(from_info, to_info, name)
 
@@ -578,10 +708,10 @@ class LocalRemote(BaseRemote):
         )
 
     @staticmethod
-    def _log_missing_caches(checksum_info_dict):
+    def _log_missing_caches(hash_info_dict):
         missing_caches = [
             (md5, info)
-            for md5, info in checksum_info_dict.items()
+            for md5, info in hash_info_dict.items()
             if info["status"] == STATUS_MISSING
         ]
         if missing_caches:
@@ -594,131 +724,3 @@ class LocalRemote(BaseRemote):
                 "nor on remote. Missing cache files: {}".format(missing_desc)
             )
             logger.warning(msg)
-
-    def _unprotect_file(self, path):
-        if System.is_symlink(path) or System.is_hardlink(path):
-            logger.debug("Unprotecting '{}'".format(path))
-            tmp = os.path.join(os.path.dirname(path), "." + uuid())
-
-            # The operations order is important here - if some application
-            # would access the file during the process of copyfile then it
-            # would get only the part of file. So, at first, the file should be
-            # copied with the temporary name, and then original file should be
-            # replaced by new.
-            copyfile(path, tmp, name="Unprotecting '{}'".format(relpath(path)))
-            remove(path)
-            os.rename(tmp, path)
-
-        else:
-            logger.debug(
-                "Skipping copying for '{}', since it is not "
-                "a symlink or a hardlink.".format(path)
-            )
-
-        os.chmod(path, self._file_mode)
-
-    def _unprotect_dir(self, path):
-        assert is_working_tree(self.repo.tree)
-
-        for fname in self.repo.tree.walk_files(path):
-            self._unprotect_file(fname)
-
-    def unprotect(self, path_info):
-        path = path_info.fspath
-        if not os.path.exists(path):
-            raise DvcException(
-                "can't unprotect non-existing data '{}'".format(path)
-            )
-
-        if os.path.isdir(path):
-            self._unprotect_dir(path)
-        else:
-            self._unprotect_file(path)
-
-    def protect(self, path_info):
-        path = fspath_py35(path_info)
-        mode = self.CACHE_MODE
-
-        try:
-            os.chmod(path, mode)
-        except OSError as exc:
-            # There is nothing we need to do in case of a read-only file system
-            if exc.errno == errno.EROFS:
-                return
-
-            # In shared cache scenario, we might not own the cache file, so we
-            # need to check if cache file is already protected.
-            if exc.errno not in [errno.EPERM, errno.EACCES]:
-                raise
-
-            actual = stat.S_IMODE(os.stat(path).st_mode)
-            if actual != mode:
-                raise
-
-    def _get_unpacked_dir_path_info(self, checksum):
-        info = self.checksum_to_path_info(checksum)
-        return info.with_name(info.name + self.UNPACKED_DIR_SUFFIX)
-
-    def _remove_unpacked_dir(self, checksum):
-        path_info = self._get_unpacked_dir_path_info(checksum)
-        self.remove(path_info)
-
-    def _path_info_changed(self, path_info):
-        if self.exists(path_info) and self.state.get(path_info):
-            return False
-        return True
-
-    def _update_unpacked_dir(self, checksum):
-        unpacked_dir_info = self._get_unpacked_dir_path_info(checksum)
-
-        if not self._path_info_changed(unpacked_dir_info):
-            return
-
-        self.remove(unpacked_dir_info)
-
-        try:
-            dir_info = self.get_dir_cache(checksum)
-            self._create_unpacked_dir(checksum, dir_info, unpacked_dir_info)
-        except DvcException:
-            logger.warning("Could not create '{}'".format(unpacked_dir_info))
-
-            self.remove(unpacked_dir_info)
-
-    def _create_unpacked_dir(self, checksum, dir_info, unpacked_dir_info):
-        self.makedirs(unpacked_dir_info)
-
-        for entry in Tqdm(dir_info, desc="Creating unpacked dir", unit="file"):
-            entry_cache_info = self.checksum_to_path_info(
-                entry[self.PARAM_CHECKSUM]
-            )
-            relative_path = entry[self.PARAM_RELPATH]
-            # In shared cache mode some cache files might not be owned by the
-            # user, so we need to use symlinks because, unless
-            # /proc/sys/fs/protected_hardlinks is disabled, the user is not
-            # allowed to create hardlinks to files that he doesn't own.
-            link_types = ["hardlink", "symlink"]
-            self._link(
-                entry_cache_info, unpacked_dir_info / relative_path, link_types
-            )
-
-        self.state.save(unpacked_dir_info, checksum)
-
-    def _changed_unpacked_dir(self, checksum):
-        status_unpacked_dir_info = self._get_unpacked_dir_path_info(checksum)
-
-        return not self.state.get(status_unpacked_dir_info)
-
-    def _get_unpacked_dir_names(self, checksums):
-        unpacked = set()
-        for c in checksums:
-            if self.is_dir_checksum(c):
-                unpacked.add(c + self.UNPACKED_DIR_SUFFIX)
-        return unpacked
-
-    def is_protected(self, path_info):
-        if not self.exists(path_info):
-            return False
-
-        mode = os.stat(fspath_py35(path_info)).st_mode
-
-        return stat.S_IMODE(mode) == self.CACHE_MODE
